import isgb

import grpc

from dataclasses import dataclass
from contextlib import contextmanager
from sgb.consts.rpc import RPC as CONST_RPC
from sgb.collections import nint, nstr, nbool, Any, Callable
from sgb.tools.service import (
    ServiceTool,
    IRootSupporter,
    ServiceRoleTool,
    IServiceCollection,
)
from sgb.tools import (
    e,
    j,
    n,
    nl,
    nn,
    ne,
    j_p,
    OSTool,
    EnumTool,
    DataTool,
    FormatTool,
    BitMask as BM,
    ParameterList,
    ErrorableThreadPoolExecutor,
)

from sgb.service.collections import (
    Subscribtion,
    ServiceDescription,
    ServiceInformation,
    SubscribtionResult,
    SubscriberInformation,
    ServiceDescriptionBase,
    SubscribtionInformation,
)
from sgb.consts.errors import Error, ServiceIsNotStartedError
from sgb.consts import SUBSCRIBTION_TYPE
import sgb.rpc.rpcCommandCall_pb2 as pb2
from sgb.consts.date_time import DATE_TIME
import sgb.rpc.rpcCommandCall_pb2_grpc as pb2_grpc
from sgb.service.consts import SERVICE_COMMAND as SC, SERVICE_ROLE

import grpc
import psutil
import socket
import atexit
import asyncio
from grpc import Server
from grpc import StatusCode
from threading import Thread
from typing import Any, Callable
from collections import defaultdict
from datetime import datetime, timedelta
from zeroconf import ServiceInfo, Zeroconf
from zeroconf import Zeroconf, ServiceBrowser, ServiceListener

DEBUG: bool = False

STUB: ServiceDescription = ServiceDescription(name="Stub")

USE_MDNS_SERVICE_DISCOVERING_DEFAULT_VALUE: bool = True


def cleanup():
    mDNS.close()


atexit.register(cleanup)


class DH:
    ROOT: IRootSupporter | None = None


def GRPC_OPTIONS() -> list[tuple[str, int]]:
    size: int = 20 * 1024 * 1024
    return [
        ("grpc.max_receive_message_length", size),
        ("grpc.max_send_message_length", size),
    ]


class IService:

    def serve(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
        call_handler: Callable[[SC | str, ParameterList, Any], Any] | None = None,
        starts_handler: Callable[[Any], None] | Callable[[], None] | None = None,
        max_workers: nint = None,
        as_standalone: bool = False,
        show_output: bool = True,
        admin_service_available: bool = True,
        use_mdns: nbool = None,
    ) -> None:
        raise NotImplemented()

    def create_subscribtions(
        self, information: ServiceDescriptionBase | None = None
    ) -> None:
        raise NotImplemented()

    @property
    def subscribtions(self) -> list[SubscribtionInformation]:
        raise NotImplemented()

    def stop(self) -> Server:
        raise NotImplemented()

    def subscribe_on(
        self,
        service_object: ServiceDescriptionBase | SERVICE_ROLE | None,
        service_command: SC | str,
        type: int = SUBSCRIBTION_TYPE.ON_RESULT,
        name: nstr = None,
    ) -> bool:
        raise NotImplemented()

    def unsubscribe(
        self, commnad_list: list[SC | str] | None = None, type: nint = None
    ) -> bool:
        raise NotImplemented()

    @property
    def description(self) -> ServiceDescription:
        raise NotImplemented()

    @property
    def information(self) -> ServiceInformation:
        raise NotImplemented()


@dataclass
class SessionInformation:
    start_time: datetime | None = None
    life_time: timedelta | None = None


class mDNSServiceListener(ServiceListener):

    def __init__(self):
        self.block = asyncio.Event()
        self.found_service: dict = None

    def add_service(self, zeroconfig: Zeroconf, service_type: str, name: str) -> None:
        if not self.block.is_set():
            info = zeroconfig.get_service_info(service_type, name)
            if info:
                self.found_service = {
                    "name": name,
                    "ip": socket.inet_ntoa(info.addresses[0]),
                    "port": info.port,
                    "type": service_type,
                    "properties": info.decoded_properties,
                }
                self.block.set()


class mDNS:

    zeroconf: Zeroconf | None = None

    @staticmethod
    def find_service(
        service_description: ServiceDescription,
    ) -> ServiceDescription | None:

        async def discover_first_service(service_type, timeout=10):
            zeroconf = Zeroconf()
            listener: mDNSServiceListener = mDNSServiceListener()
            browser: ServiceBrowser = ServiceBrowser(zeroconf, service_type, listener)
            try:
                await asyncio.wait_for(listener.block.wait(), timeout)
                return listener.found_service
            except asyncio.TimeoutError:
                return None
            finally:
                zeroconf.close()

        async def discover_service(
            service_description: ServiceDescription,
        ) -> ServiceDescription | None:
            service_role: SERVICE_ROLE = ServiceRoleTool.get(service_description)
            service_type: str = j_p(
                (
                    j(
                        (
                            "_",
                            OSTool.domain(),
                            "-",
                            service_role.name.lower(),
                        )
                    ),
                    "_tcp",
                    "local",
                    "",
                )
            )
            service = await discover_first_service(service_type)
            if service:
                return DataTool.fill_data_from_source(
                    ServiceDescription(),
                    service["properties"],
                    skip_not_none=True,
                )
            else:
                return None

        return asyncio.run(discover_service(service_description))

    @staticmethod
    def close() -> None:
        if nn(mDNS.zeroconf):
            mDNS.zeroconf.unregister_all_services()
            mDNS.zeroconf.close()

    @staticmethod
    def register_service(service_description: ServiceDescription) -> None:
        service_type: str = j_p(
            (
                j(
                    (
                        "_",
                        OSTool.domain(),
                        "-",
                        service_description.standalone_name,
                    )
                ),
                "_tcp",
                "local",
                "",
            )
        )
        service_info = ServiceInfo(
            service_type,
            j_p(
                (
                    service_description.description,
                    service_type,
                )
            ),
            addresses=[socket.inet_aton(OSTool.host_ip())],
            port=service_description.port,
            properties=DataTool.to_data(service_description) or {},
            server=OSTool.host(),
        )

        # Регистрируем сервис
        mDNS.zeroconf = Zeroconf()
        mDNS.zeroconf.register_service(service_info)


class ServiceClient:

    server: Server | None = None

    context: grpc.RpcContext | None = None

    service: IService | None = None

    service_description: ServiceDescription | None = None

    service_information: ServiceInformation | None = None

    session_information: SessionInformation = SessionInformation()

    def __init__(
        self,
        root: IRootSupporter | None = None,
        use_mdns_service_discovering: bool = USE_MDNS_SERVICE_DISCOVERING_DEFAULT_VALUE,
    ):
        self.service_collection: IServiceCollection = ServiceCollection()
        self.use_mdns_service_discovering = use_mdns_service_discovering
        if nn(root):
            DH.ROOT = root
        self.admin_service_available: nbool = None

    
    def create_error(self, context, message: str = "", code: Any = None) -> Any:
        context.set_details(message)
        context.set_code(code)
        return pb2.rpcCommandResult()

    def get_host(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> nstr:
        service_description: ServiceDescriptionBase | None = (
            ServiceRoleTool.service_description(service_object)
        )
        if isinstance(service_description, ServiceDescription):
            if ne(service_description.port):
                return FormatTool.domain(OSTool.domain_dns(), service_description.host)
        value: nstr = (
            self.get_information(service_object) or ServiceDescriptionBase()
        ).host or service_description.host
        if n(value):
            return value
        return value  # FormatTool.domain(DH.ROOT.domain_dns, value)

    
    def check_on_availability(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
        cached: bool = False,
    ) -> bool:
        return ne(self.get_information(service_object, cached))

    def get_information_from_cache(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> ServiceInformation | None:
        return self.service_collection.get(
            ServiceRoleTool.service_description(service_object)
        )

    def get_port(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> nint:
        service_description: ServiceDescriptionBase | None = (
            ServiceRoleTool.service_description(service_object)
        )
        if isinstance(service_description, ServiceDescription):
            if ne(service_description.port):
                return service_description.port
        return (
            self.get_information(service_object) or ServiceDescriptionBase()
        ).port or service_description.port

    def get_information(
        self, service_object: SERVICE_ROLE | ServiceDescriptionBase, cached: bool = True
    ) -> ServiceInformation | None:
        service_description: ServiceDescriptionBase | None = (
            ServiceRoleTool.service_description(service_object)
        )
        if e(self.service_collection.get()):
            self.service_collection.request(self)
        service_description = self.get_information_from_cache(service_description)
        if cached:
            return service_description
        # if e(service_description):
        #    return None
        return self.ping(service_object)

    def get_service_information_list(self) -> list[ServiceInformation]:
        return DataTool.map(
            ServiceTool.extract_service_information,
            DataTool.rpc_decode(
                self.call_service(
                    None,
                    SC.get_service_information_list,
                    ((self.service_information or STUB).name),
                )
            )
            or [],
        )

    def ping(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> ServiceInformation | None:
        try:
            service_information: ServiceDescriptionBase = (
                ServiceRoleTool.service_description(service_object)
            )
            value = DataTool.fill_data_from_rpc_str(
                ServiceInformation(),
                self.call_service(
                    service_information,
                    SC.ping,
                    ((self.service_information or STUB).name),
                ),
            )
            return value
        except Error:
            return None

    def call_service_command(
        self, command: SC | str, parameters: Any = None, timeout: nint = None
    ) -> nstr:
        return self.call_service(None, command, parameters, timeout)

    def call_service(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase | None,
        command: SC | str,
        parameters: Any = None,
        timeout: nint = None,
    ) -> nstr:

        service_description: ServiceDescriptionBase = (
            (ServiceRoleTool.service_description(service_object))
            if nn(service_object)
            else DH.ROOT.description_by_service_command(command)
        )
        try:
            service_host: nstr = self.get_host(service_description)
            if n(service_host):
                if self.use_mdns_service_discovering:
                    mdns_service_description: ServiceDescription | None = (
                        mDNS.find_service(service_description)
                    )
                    if nn(mdns_service_description):
                        service_description = mdns_service_description
                        self.service_collection.update(service_description, add=True)
                        service_host = service_description.host
                if n(service_host):
                    error: ServiceIsNotStartedError = ServiceIsNotStartedError(
                        "Service is not started"
                    )
                    code: StatusCode = StatusCode.UNAVAILABLE
                    details: str = j(
                        (
                            "Service role name: ",
                            service_description.name,
                            nl(),
                            "Command: ",
                            command,
                            nl(),
                            "Details: ",
                            error.details,
                            nl(),
                            "Code: ",
                            code,
                        )
                    )
                    DH.ROOT.error_handler(
                        service_description, error, code, details, command
                    )
                    raise error
            service_port: nint = self.get_port(service_description)
            if e(timeout):
                if (
                    e(self.service_information)
                    or self.service_information.isolated
                    or service_description.isolated
                ):
                    if command == SC.ping:
                        timeout = CONST_RPC.TIMEOUT_FOR_PING
                    else:
                        timeout = CONST_RPC.TIMEOUT
            return (
                CommandClient(service_host, service_port)
                .call_command(
                    command.name if isinstance(command, SC) else command,
                    DataTool.rpc_encode(parameters),
                    timeout,
                )
                .data
            )
        except grpc.RpcError as error:
            code: tuple = error.code()
            details: str = j(
                (
                    "Service role name: ",
                    service_description.name,
                    nl(),
                    "Host: ",
                    service_host,
                    ":",
                    service_port,
                    nl(),
                    "Command: ",
                    command,
                    nl(),
                    "Details: ",
                    error.details(),
                    nl(),
                    "Code: ",
                    code,
                )
            )
            DH.ROOT.error_handler(service_description, error, code, details, command)
            # raise Error(error.details(), code)

    def create_service(self) -> IService:
        return Service(self)


class UnaryServicer(pb2_grpc.UnaryServicer):
    def __init__(
        self,
        service: IService,
        client: ServiceClient,
        description: ServiceDescription,
        call_handler: Callable[[str, ParameterList, Any], Any] | None = None,
        *args,
        **kwargs,
    ):
        self.service_information = description
        self.call_handler = call_handler
        self.service = service
        self.client = client
        self.subscriber_information_map: dict[str, dict[str, SubscriberInformation]] = (
            defaultdict(dict)
        )

    def internal_call_handler(self, sc: SC | str, parameters: Any, context) -> Any:
        try:
            parameter_list: ParameterList = ParameterList(parameters)
            if ne(sc):
                if sc == SC.stop_service:
                    is_subscriber: bool = ServiceTool.is_service_as_listener(
                        self.service_information
                    )
                    if not is_subscriber:
                        self.client.call_service(
                            SC.on_service_stops, self.service_information
                        )
                    self.service.stop()
                    if not is_subscriber:
                        pid: int = OSTool.pid()
                        parent = psutil.Process(pid)
                        for child in parent.children(recursive=True):
                            child.kill()
                    return
                if sc == SC.ping:
                    return DataTool.fill_data_from_source(
                        ServiceInformation(), self.service_information
                    )
                if sc == SC.subscribe:
                    return self.subscribe_with_parameter_list(parameter_list)
                if sc == SC.unsubscribe:
                    return self.unsubscribe_with_parameter_list(parameter_list)
                if sc == SC.create_subscribtions:
                    self.service.create_subscribtions(
                        parameter_list.next(ServiceDescriptionBase())
                    )
                    return True
                if sc == SC.update_service_information:
                    self.client.service_collection.update(
                        parameter_list.next_as_list(ServiceInformation),
                        parameter_list.next(),
                    )
                    return True
                if sc == SC.heart_beat:
                    date_string: str = ServiceTool.extract_parameter_list(
                        parameter_list
                    ).get()
                    date: datetime = datetime.strptime(
                        date_string, DATE_TIME.ISO_DATETIME_FORMAT
                    )
                    parameter_list.set(0, date)
                    self.client.session_information.life_time = (
                        date - self.client.session_information.start_time
                    )
            return (
                None
                if e(self.call_handler)
                else self.call_subscribers_after(
                    sc,
                    parameter_list,
                    self.call_handler(
                        sc,
                        self.call_subscribers_before(sc, parameter_list),
                        context,
                    ),
                )
            )
        except Exception as exception:
            DH.ROOT.error_detect_handler(exception)

    def unsubscribe_all(self) -> None:

        def unsubscribe_all_internal() -> None:
            for sc in self.subscriber_information_map:
                for service_information_name in self.subscriber_information_map[sc]:
                    subscriber_information: SubscriberInformation = (
                        self.subscriber_information_map[sc][service_information_name]
                    )
                    DataTool.rpc_decode(
                        self.client.call_service(
                            subscriber_information.service_information,
                            SC.unsubscribe,
                            (sc, subscriber_information.name),
                        )
                    )
            self.subscriber_information_map = defaultdict(dict)

        Thread(target=unsubscribe_all_internal).start()

    def unsubscribe_with_parameter_list(self, value: ParameterList) -> None:
        description: ServiceDescription = value.next(ServiceDescription())
        role_description_name: str = description.name
        for service_command in self.subscriber_information_map:
            if (
                role_description_name
                in self.subscriber_information_map[service_command]
            ):
                del self.subscriber_information_map[service_command][
                    role_description_name
                ]

    def subscribe_with_parameter_list(self, pl: ParameterList) -> bool:
        service_information: ServiceDescriptionBase = pl.next(ServiceDescriptionBase())
        subscribtion_information: SubscribtionInformation = pl.next(
            SubscribtionInformation()
        )
        subscribtion_information.service_command = EnumTool.get(
            SC, subscribtion_information.service_command
        ) or subscribtion_information.service_command
        return self.subscribe(service_information, subscribtion_information)

    def subscribe(
        self,
        service_description: ServiceDescription,
        subscribtion_information: SubscribtionInformation,
    ) -> bool:
        sc: SC | str = subscribtion_information.service_command
        name: nstr = subscribtion_information.name
        type: int = subscribtion_information.type
        if sc in self.subscriber_information_map:
            if service_description.name in self.subscriber_information_map[sc]:
                subscriber_information: SubscriberInformation = (
                    self.subscriber_information_map[sc][service_description.name]
                )
                if (
                    subscriber_information.service_information == service_description
                    and BM.has(subscriber_information.type, type)
                ):
                    self.subscriber_information_map[sc][service_description.name] = (
                        SubscriberInformation(
                            type, name, service_information=service_description
                        )
                    )
            else:
                self.subscriber_information_map[sc][service_description.name] = (
                    SubscriberInformation(
                        type, name, service_information=service_description
                    )
                )
        else:
            self.subscriber_information_map[sc][service_description.name] = (
                SubscriberInformation(
                    type, name, service_information=service_description
                )
            )
        return True

    def call_subscribers_before(self, sc: SC | str, in_result: ParameterList):
        out_result: ParameterList = in_result
        if sc in self.subscriber_information_map:
            for service_information_name in list(self.subscriber_information_map[sc]):
                subscriber_information: SubscriberInformation = (
                    self.subscriber_information_map[sc][service_information_name]
                )
                service_information: ServiceInformation = (
                    subscriber_information.service_information
                )
                if (
                    BM.has(subscriber_information.type, SUBSCRIBTION_TYPE.ON_CALL)
                    and subscriber_information.enabled
                ):
                    subscriber_information.available = self.client.check_on_availability(
                        service_information
                    )
                    if subscriber_information.available:
                        out_result = ParameterList(
                            DataTool.rpc_decode(
                                self.client.call_service(
                                    service_information,
                                    sc,
                                    (
                                        in_result,
                                        SubscribtionResult(
                                            None,
                                            SUBSCRIBTION_TYPE.ON_CALL,
                                            True,
                                        ),
                                    ),
                                )
                            )
                        )
                    else:
                        del self.subscriber_information_map[sc][
                            service_information_name
                        ]
        return out_result

    def call_subscribers_after(
        self, sc: SC | str, parameter_list: ParameterList, result: Any
    ) -> Any:

        def internal_call_subscribers_after_sequentially(
            command: SC | str,
            subscriber_list: list[SubscriberInformation],
            parameter_list: ParameterList,
            result: Any,
        ):
            for subscriber in subscriber_list:
                role_description_item: ServiceDescription = (
                    subscriber.service_information
                )
                subscriber.available = self.client.check_on_availability(
                    role_description_item
                )
                if subscriber.available:
                    result = self.client.call_service(
                        role_description_item,
                        command,
                        (
                            parameter_list,
                            SubscribtionResult(
                                result,
                                SUBSCRIBTION_TYPE.ON_RESULT_SEQUENTIALLY,
                                True,
                            ),
                        ),
                    )
                else:
                    if DataTool.if_is_in(
                        self.subscriber_information_map[command],
                        role_description_item.name,
                    ):
                        del self.subscriber_information_map[command][
                            role_description_item.name
                        ]

        def internal_call_subscribers_after(
            sc: SC | str,
            subscriber_information: SubscriberInformation,
            parameter_list: ParameterList,
            result: Any,
        ):
            service_information: ServiceDescriptionBase = (
                subscriber_information.service_information
            )
            subscriber_information.available = self.client.check_on_availability(
                service_information
            )
            if subscriber_information.available:
                self.client.call_service(
                    service_information,
                    sc,
                    (
                        parameter_list,
                        SubscribtionResult(result, SUBSCRIBTION_TYPE.ON_RESULT, True),
                    ),
                )
            else:
                if service_information.name in self.subscriber_information_map[sc]:
                    del self.subscriber_information_map[sc][service_information.name]

        if sc in self.subscriber_information_map:
            after_sequentially_subscriber_information_list: list[
                SubscriberInformation
            ] = []
            for servce_information_name in list(self.subscriber_information_map[sc]):
                subscriber_information: SubscriberInformation = (
                    self.subscriber_information_map[sc][servce_information_name]
                )
                if subscriber_information.enabled:
                    if BM.has(subscriber_information.type, SUBSCRIBTION_TYPE.ON_RESULT):
                        Thread(
                            target=internal_call_subscribers_after,
                            args=(
                                sc,
                                subscriber_information,
                                parameter_list,
                                result,
                            ),
                        ).start()
                    elif BM.has(
                        subscriber_information.type,
                        SUBSCRIBTION_TYPE.ON_RESULT_SEQUENTIALLY,
                    ):
                        after_sequentially_subscriber_information_list.append(
                            subscriber_information
                        )
            if ne(after_sequentially_subscriber_information_list):
                Thread(
                    target=internal_call_subscribers_after_sequentially,
                    args=(
                        sc,
                        after_sequentially_subscriber_information_list,
                        parameter_list,
                        result,
                    ),
                ).start()

        return result

    def rpcCallCommand(self, command_object, context):

        parameters: Any = command_object.parameters

        if ne(parameters):
            parameters = DataTool.rpc_decode(parameters)
        ServiceClient.context = context

        command: SC | nstr = EnumTool.get(SC, command_object.name)
        if n(command):
            command = command_object.name
        result: Any = self.internal_call_handler(command, parameters, context)
        if n(context.code()):
            return pb2.rpcCommandResult(data=DataTool.rpc_encode(result))
        return result


class CommandClient:
    def __init__(self, host: str, port: int):
        self.stub = pb2_grpc.UnaryStub(
            grpc.insecure_channel(
                j((host, port), ":"),
                options=GRPC_OPTIONS(),
                # compression=grpc.Compression.Gzip,
            )
        )

    def call_command(self, name: str, parameters: nstr = None, timeout: nint = None):
        return self.stub.rpcCallCommand(
            pb2.rpcCommand(name=name, parameters=parameters), timeout=timeout
        )


class ServiceCollection(IServiceCollection):

    service_collection: dict[ServiceDescriptionBase, ServiceInformation] = {}

    def request(self, client: ServiceClient) -> None:
        self.update(
            client.get_service_information_list(),
            True,
            True,
        )

    def remove_service(self, value: ServiceDescriptionBase) -> None:
        if value in self.service_collection:
            del self.service_collection[value]

    def clear(self) -> None:
        self.service_collection = {}

    def get(
        self,
        key_description: ServiceDescription | None = None,
    ) -> dict[ServiceDescriptionBase, ServiceInformation] | ServiceInformation | None:
        return (
            (
                self.service_collection[key_description]
                if key_description in self.service_collection
                else None
            )
            if nn(key_description)
            else self.service_collection
        )

    def update(
        self,
        value: ServiceDescriptionBase | list[ServiceDescriptionBase],
        add: bool = True,
        overwrite: bool = False,
    ) -> None:
        if not isinstance(value, list):
            value = [value]
        if overwrite:
            if ne(value):
                self.service_collection = {}
        for item in value:
            if add:
                self.service_collection[item] = item
            else:
                if item in self.service_collection:
                    del self.service_collection[item]


class Service(IService):

    MAX_WORKERS: int = 10

    def __init__(self, client: ServiceClient) -> None:
        self.client = client
        self._description: ServiceDescription | None = None
        self._information: ServiceInformation | None = None
        self.subscribtions_map: dict[
            ServiceDescriptionBase, dict[str, Subscribtion]
        ] = defaultdict(dict)
        self.server: Server | None = None

    @property
    def description(self) -> ServiceDescription:
        return self._description

    @property
    def information(self) -> ServiceInformation:
        return self._information

    @contextmanager
    def detect_error(self):
        try:
            yield True
        except Exception as exception:
            DH.ROOT.error_detect_handler(exception)
        finally:
            pass

    def serve(
        self,
        service_object: SERVICE_ROLE | ServiceDescriptionBase,
        call_handler: Callable[[SC | str, ParameterList, Any], Any] | None = None,
        starts_handler: Callable[[IService], None] | Callable[[], None] | None = None,
        max_workers: nint = None,
        as_standalone: bool = False,
        show_output: bool = True,
        admin_service_available: bool = True,
        use_mdns: nbool = None,
    ) -> None:
        self.client.admin_service_available = admin_service_available
        max_workers = max_workers or Service.MAX_WORKERS
        service_description: ServiceDescription = (
            service_object
            if isinstance(service_object, ServiceDescriptionBase)
            else ServiceRoleTool.service_description(service_object)
        )
        self._description = service_description
        with self.detect_error():
            if not ServiceTool.is_service_as_listener(service_description):
                DH.ROOT.add_isolated_arg()
                isolate_arg: nstr = None
                try:
                    isolated_arg_value: nstr = DH.ROOT.isolated_arg()
                    if nn(isolated_arg_value):
                        isolate_arg = str(isolated_arg_value).lower()
                except AttributeError as error:
                    pass
                if nn(isolate_arg):
                    if n(self.description.isolated):
                        self.description.isolated = isolate_arg not in [
                            "0",
                            "no",
                            "false",
                        ]
                    elif self.description.isolated:
                        self.description.isolated = isolate_arg not in [
                            "0",
                            "no",
                            "false",
                        ]
        self.description.host = ServiceTool.create_host(self.description)
        self.description.port = ServiceTool.create_port(self.description)
        self._information = DataTool.fill_data_from_source(
            ServiceInformation(standalone=as_standalone),
            self.description,
            skip_not_none=True,
        )
        self.information.pid = OSTool.pid()
        if e(self.client.service):
            self.client.service_description = self.description
            self.client.service_information = self.information
            self.client.service = self
        self.server = grpc.server(
            ErrorableThreadPoolExecutor(max_workers=max_workers),
            options=GRPC_OPTIONS(),
            # compression=grpc.Compression.Gzip,
        )
        if show_output:
            DH.ROOT.service_header(self.information)
        pb2_grpc.add_UnaryServicer_to_server(
            UnaryServicer(self, self.client, self.description, call_handler),
            self.server,
        )
        try:
            self.server.add_insecure_port(
                j((self.description.host, self.description.port), ":")
            )

            self.server.start()
            self.client.session_information.start_time = datetime.now().replace(
                second=0, microsecond=0
            )
            if nn(starts_handler):
                starts_handler(self)
            self.information.subscribtions = self.subscribtions
            if admin_service_available:
                self.client.call_service_command(
                    SC.on_service_starts, (self.information,)
                )
            else:
                self.create_subscribtions()
            if not admin_service_available or use_mdns:
                mDNS.register_service(self.description)
            self.server.wait_for_termination()
        except RuntimeError as error:
            DH.ROOT.service_was_not_started(self.description, j(error.args))

    def create_subscribtions(
        self, information: ServiceDescriptionBase | None = None
    ) -> None:
        def internal_create_subscribtion(
            subscribtions: dict[ServiceDescriptionBase, dict[SC | str, Subscribtion]],
            information: ServiceDescriptionBase,
        ):
            for sc in subscribtions[information]:
                subscription: Subscribtion = subscribtions[information][sc]
                self.client.call_service(
                    information,
                    SC.subscribe,
                    (
                        DataTool.fill_data_from_source(
                            ServiceDescriptionBase(), self.information
                        ),
                        DataTool.fill_data_from_source(
                            SubscribtionInformation(), subscription
                        ),
                    ),
                )
                subscription.available = True
                subscription.enabled = True

        for description_item in (
            self.subscribtions_map if e(information) else [information]
        ):
            if self.client.check_on_availability(description_item):
                internal_create_subscribtion(self.subscribtions_map, description_item)

    @property
    def subscribtions(self) -> list[Subscribtion]:
        result: list[Subscribtion] = []
        for description_item in self.subscribtions_map:
            for sc in self.subscribtions_map[description_item]:
                result.append(self.subscribtions_map[description_item][sc])
        return result

    def subscribe_on(
        self,
        service_object: ServiceDescriptionBase | SERVICE_ROLE | None,
        service_command: SC | str,
        type: int = SUBSCRIBTION_TYPE.ON_RESULT,
        name: nstr = None,
    ) -> bool:
        service_description: ServiceDescription | None = (ServiceRoleTool.service_description(service_object) if nn(service_object) else None) or (
            DH.ROOT.description_by_service_command(service_command)
        ) or ServiceRoleTool.service_description(SERVICE_ROLE.ROUTER)
        if ne(service_description):
            if service_description != self.description:
                subscribtion: Subscribtion | None = None
                if service_command not in self.subscribtions_map[service_description]:
                    subscribtion = Subscribtion(service_description.name, service_command, type, name)
                    self.subscribtions_map[service_description][
                        service_command
                    ] = subscribtion
                else:
                    subscribtion = self.subscribtions_map[service_description][
                        service_command
                    ]
                    subscribtion.type |= type
                subscribtion.available = self.client.check_on_availability(
                    service_description
                )
                if self.client.admin_service_available:
                    return subscribtion.available
                else:
                    return subscribtion.available
        return False

    def unsubscribe(
        self, command_list: list[SC | str] | None = None, type: nint = None
    ) -> None:

        for service_description in self.subscribtions_map:
            for sc in self.subscribtions_map[service_description]:
                if sc in command_list:
                    DataTool.rpc_decode(
                        self.client.call_service(
                            service_description, "unsubscribe", (self.description,)
                        )
                    )

    def unsubscribe_all(
        self, service_object: SERVICE_ROLE | ServiceDescriptionBase
    ) -> bool:
        pass
        # return self.unsubscribe(service_object)

    def stop(self) -> None:
        self.server.stop(0)



import isgb

from sgb.tools import j, j_p
from sgb.collections import strtuple, strlist
from sgb.consts.names import ACTIVE_DIRECTORY_USER_PROPERTIES

from enum import Enum


class AD:

    SPLITTER: str = "."
    OU: str = "OU="
    SEARCH_ATTRIBUTES: strlist = [
        ACTIVE_DIRECTORY_USER_PROPERTIES.LOGIN,
        ACTIVE_DIRECTORY_USER_PROPERTIES.NAME,
    ]
    SEARCH_ATTRIBUTE_DEFAULT: str = SEARCH_ATTRIBUTES[0]
    SEARCH_ALL_PATTERN: str = "*"
    LOCATION_SPLITTER: str = ":"
    TEMPLATED_USER_SERACH_TEMPLATE: str = j(("_", SEARCH_ALL_PATTERN, "_"))

    USER_ACCOUNT_CONTROL: strlist = [
        "SCRIPT",
        "ACCOUNTDISABLE",
        "RESERVED",
        "HOMEDIR_REQUIRED",
        "LOCKOUT",
        "PASSWD_NOTREQD",
        "PASSWD_CANT_CHANGE",
        "ENCRYPTED_TEXT_PWD_ALLOWED",
        "TEMP_DUPLICATE_ACCOUNT",
        "NORMAL_ACCOUNT",
        "RESERVED",
        "INTERDOMAIN_TRUST_ACCOUNT",
        "WORKSTATION_TRUST_ACCOUNT",
        "SERVER_TRUST_ACCOUNT",
        "RESERVED",
        "RESERVED",
        "DONT_EXPIRE_PASSWORD",
        "MNS_LOGON_ACCOUNT",
        "SMARTCARD_REQUIRED",
        "TRUSTED_FOR_DELEGATION",
        "NOT_DELEGATED",
        "USE_DES_KEY_ONLY",
        "DONT_REQ_PREAUTH",
        "PASSWORD_EXPIRED",
        "TRUSTED_TO_AUTH_FOR_DELEGATION",
        "RESERVED",
        "PARTIAL_SECRETS_ACCOUNT",
    ]

    DOMAIN_NAME: str = "sgb"
    DOMAIN_SUFFIX: str = "lan"
    DOMAIN_DNS: str = j_p((DOMAIN_NAME, DOMAIN_SUFFIX))
    #DOMAIN: str = DOMAIN_DNS
    #PATH_ROOT: str = j(("//", DOMAIN))

    ROOT_CONTAINER_DN: str = f"DC={DOMAIN_NAME},DC={DOMAIN_SUFFIX}"
    WORKSTATIONS_CONTAINER_DN: str = f"{OU}Workstations,{ROOT_CONTAINER_DN}"
    SERVERS_CONTAINER_DN: str = f"{OU}Servers,{ROOT_CONTAINER_DN}"
    ADMINS_CONTAINER_DN: str = f"{OU}Admins,{ROOT_CONTAINER_DN}"
    USERS_CONTAINER_DN_SUFFIX: str = f"Holding,{ROOT_CONTAINER_DN}"
    OUTSOURCE_USERS_CONTAINER_DN_SUFFIX: str = f"Outsourcing,{ROOT_CONTAINER_DN}"
    ACTIVE_USERS_CONTAINER_DN: strtuple = f"{OU}{USERS_CONTAINER_DN_SUFFIX}"
    OUTSOURCE_USERS_CONTAINER_DN: str = f"{OU}{OUTSOURCE_USERS_CONTAINER_DN_SUFFIX}"
    ALL_ACTIVE_USERS_CONTAINER_DN: strtuple = (
        ACTIVE_USERS_CONTAINER_DN,
        OUTSOURCE_USERS_CONTAINER_DN,
        ADMINS_CONTAINER_DN,
    )
    INACTIVE_USERS_CONTAINER_DN: str = f"{OU}User,{OU}Disabled,{ROOT_CONTAINER_DN}"
    GROUP_CONTAINER_DN: str = f"{OU}Groups,{ROOT_CONTAINER_DN}"
    PROPERTY_ROOT_DN: str = f"{OU}Property,{GROUP_CONTAINER_DN}"
    # PROPERTY_COMPUTER_DN: str = f"{OU}Computer,{PROPERTY_ROOT_DN}"
    PROPERTY_USER_DN: str = f"{OU}User,{PROPERTY_ROOT_DN}"
    JOB_POSITION_CONTAINER_DN: str = f"{OU}Job positions,{GROUP_CONTAINER_DN}"

    class ComputerProperties(Enum):
        Watchable = 1
        Shutdownable = 2
        Rebootable = 4
        DiskReportable = 8
        DiskReportableViaZabbix = 16

    class UserProperies(Enum):
        Jokeless = 1
        DoctorVisitless = 2
        HasLunchBreak = 4
        TimeTrackingless = 8



import isgb


from sgb.collections import nstr

import enum


class COMMUNITY(enum.Enum):

    SPHERE = "Сфера"

    GAS = "Грин-Агро Сахалин"

    PK = "Приморский кондитер"

    AGMZ = "Артёмоский гормолокозавод"

    GAK = "Грин Агро-Кировский"

    HAPK = "Грин-Агро «Ханкайский агро-промышленный комплекс»"

    TD = "ТД «Дальневосточные Продукты»"

    MKUS = "Молочный комбинат Южно-Сахалинский"

    ANT = "Охранное агенство «АНТ»"

    DP = "СПССК 'Доступные продукты'"


class COMMUNITY_SETTING_ITEM(enum.Enum):
    SIP_TELEPHONELESS: int = 1


COMMUNITY_SETTINGS: dict[
    COMMUNITY, COMMUNITY_SETTING_ITEM | tuple[COMMUNITY_SETTING_ITEM] | None
] = {
    COMMUNITY.SPHERE: None,
    COMMUNITY.GAS: None,
    COMMUNITY.PK: None,
    COMMUNITY.AGMZ: None,
    COMMUNITY.HAPK: None,
    COMMUNITY.TD: None,
    COMMUNITY.MKUS: None,
    COMMUNITY.ANT: COMMUNITY_SETTING_ITEM.SIP_TELEPHONELESS,
    COMMUNITY.DP: COMMUNITY_SETTING_ITEM.SIP_TELEPHONELESS,
    COMMUNITY.GAK: COMMUNITY_SETTING_ITEM.SIP_TELEPHONELESS
}


COMMUNITY_USER_CONTAINER_NAME: dict[COMMUNITY, nstr] = {
    COMMUNITY.SPHERE: None,
    COMMUNITY.GAS: "greeangro",
    COMMUNITY.PK: "primkon",
    COMMUNITY.AGMZ: None,
    COMMUNITY.HAPK: None,
    COMMUNITY.TD: None,
    COMMUNITY.MKUS: None,
    COMMUNITY.ANT: None,
    COMMUNITY.DP: None,
    COMMUNITY.GAK: None
}



import isgb

from sgb.collections import ServerRoleItem
from enum import Enum

class SERVER_ROLE(Enum):

    DC: ServerRoleItem = ServerRoleItem("domain controller", "dc")
    TELEPHONY: ServerRoleItem = ServerRoleItem("asterisk", "pbx")


import os

from sgb.consts.facade import *
from sgb.consts.file import FILE
from isgb import ROOT_MODULE_NAME
from sgb.consts.python import PYTHON
from sgb.tools import j, n, PathTool, nn


from setuptools.dist import Distribution

PATH_SPLITTER: str = "/"
PATH_DOUBLE_SPLITTER: str = "//"


class PATH_FACADE:
    LINUX_MOUNT_POINT: str = j((PATH_SPLITTER, "mnt"))
    LINUX_MOUNT_POINT_PATH: str = j((LINUX_MOUNT_POINT, FACADE.NAME), PATH_SPLITTER)
    FILES_NAME: str = ".files"
    TOOLS_NAME: str = ".tools"
    SCRIPTS_NAME: str = "scripts"

    VALUE: str = "C:\\facade\\"

    """
    j(
        (
            PATH_DOUBLE_SPLITTER,
            AD.DOMAIN,
            PATH_SPLITTER,
            FACADE.NAME,
            PATH_SPLITTER,
        )
    )
    """

    @staticmethod
    def SERVICE(service_name: str) -> str:
        return os.path.join(
            PATH_FACADE.VALUE,
            FACADE.SERVICE_NAME(service_name),
        )

    @staticmethod
    def SERVICE_FILES(standalone_name: str) -> str:
        return os.path.join(PATH_FACADE.FILES(), standalone_name)

    @staticmethod
    def SERVICE_TOOLS(standalone_name: str) -> str:
        return os.path.join(PATH_FACADE.TOOLS(), standalone_name)

    @staticmethod
    def FILES() -> str:
        return os.path.join(PATH_FACADE.VALUE, PATH_FACADE.FILES_NAME)

    @staticmethod
    def TOOLS() -> str:
        return os.path.join(PATH_FACADE.VALUE, PATH_FACADE.TOOLS_NAME)


    class DITRIBUTIVE:

        SPLITTER: str = "-"
        FOLDER_NAME: str = ".distr"
        PACKAGE_FOLDER_NAME: str = "all"
        DEFAULT_PACKAGE_TAG: str = "py3-none-any"

        @staticmethod
        def NAME(value: str, version: str | None = None) -> str:
            return j(
                (ROOT_MODULE_NAME, value, None if n(version) else j(("==", version))),
                PATH_FACADE.DITRIBUTIVE.SPLITTER,
            )

        @staticmethod
        def VALUE() -> str:
            return os.path.join(PATH_FACADE.VALUE, PATH_FACADE.DITRIBUTIVE.FOLDER_NAME)

        @staticmethod
        def PACKAGE_FOLDER(name: str, version: str | None = None) -> str:
            path_list: list[str] = [PATH_FACADE.DITRIBUTIVE.VALUE(), name]
            if nn(version):
                path_list.insert(1, PATH_FACADE.DITRIBUTIVE.PACKAGE_FOLDER_NAME)
            return os.path.join(*path_list)

        @staticmethod
        def PACKAGE(name: str, version: str | None) -> str:
            def wheel_name(name: str, version: str | None) -> str:
                tag: str = ""
                dist_name: str = ""
                try:
                    dist: Distribution = Distribution(
                        attrs={"name": name, "version": version}
                    )
                    bdist_wheel_cmd = dist.get_command_obj("bdist_wheel")
                    bdist_wheel_cmd.ensure_finalized()
                    dist_name = bdist_wheel_cmd.wheel_dist_name
                    tag = PATH_FACADE.DITRIBUTIVE.SPLITTER.join(
                        bdist_wheel_cmd.get_tag()
                    )
                except BaseException as _:
                    dist_name = PATH_FACADE.DITRIBUTIVE.SPLITTER.join((name, version))
                    tag = PATH_FACADE.DITRIBUTIVE.DEFAULT_PACKAGE_TAG
                return PathTool.add_extension(
                    j((dist_name, PATH_FACADE.DITRIBUTIVE.SPLITTER, tag)),
                    PYTHON.PACKAGE_EXTENSION,
                )

            return os.path.join(
                *[
                    PATH_FACADE.DITRIBUTIVE.PACKAGE_FOLDER(name, version),
                    wheel_name(
                        j((ROOT_MODULE_NAME, name), "_"),
                        version,
                    ),
                ]
            )

    class VIRTUAL_ENVIRONMENT:

        NAME_PREFIX: str = ".venv"


class PATH_DATA_STORAGE:
    NAME: str = ".data"
    VALUE: str = os.path.join(PATH_FACADE.VALUE, NAME)


class PATH_BUILD:
    NAME: str = ".build"
    VALUE: str = os.path.join(PATH_FACADE.VALUE, NAME)


class PATH_APP:
    NAME: str = "apps"
    FOLDER: str = os.path.join(PATH_FACADE.VALUE, NAME)


class PATH_DOCS:
    NAME: str = f"Docs{FACADE.SERVICE_FOLDER_SUFFIX}"
    FOLDER: str = os.path.join(PATH_FACADE.VALUE, NAME)


class PATH_FONTS:
    NAME: str = "fonts"
    FOLDER: str = os.path.join(PATH_DOCS.FOLDER, NAME)

    @staticmethod
    def get(name: str) -> str:
        from sgb.tools import PathTool

        return os.path.join(
            PATH_FONTS.FOLDER,
            PathTool.add_extension(name, FILE.EXTENSION.TRUE_TYPE_FONT),
        )



class PATH_WS:
    NAME: str = j(("WS", FACADE.SERVICE_FOLDER_SUFFIX))
    PATH: str = os.path.join(PATH_FACADE.VALUE, NAME)


class PATHS:
    WS: PATH_WS = PATH_WS()
    DOCS: PATH_DOCS = PATH_DOCS()
    FONTS: PATH_FONTS = PATH_FONTS()
    FACADE: PATH_FACADE = PATH_FACADE()
    DATA_STORAGE: PATH_DATA_STORAGE = PATH_DATA_STORAGE()
    BUILD: PATH_BUILD = PATH_BUILD()



import isgb

from sgb.consts import CONST
from sgb.collections import nstr
from sgb.tools import (
    j,
    e,
    n,
    ne,
    OSTool,
    DataTool,
    NetworkTool,
    ParameterList,
)
from sgb.service.consts import (
    SERVICE_ROLE,
    SUPPORT_NAME_PREFIX,
    EVENT_LISTENER_NAME_PREFIX,
)


from sgb.service.collections import (
    Subscribtion,
    ServiceInformation,
    ServiceDescription,
    SubscribtionResult,
    ServiceDescriptionBase,
)

from sgb.service.consts import SERVICE_COMMAND as SC


import grpc
from enum import Enum
from abc import ABC, abstractmethod


class IRootSupporter(ABC):

    @abstractmethod
    def add_isolated_arg(self) -> bool:
        pass

    @abstractmethod
    def error_handler(
        self,
        description: ServiceDescriptionBase,
        exception: grpc.RpcError,
        code: tuple,
        details: str,
        command: str,
    ) -> None:
        pass

    @abstractmethod
    def isolated_arg(self) -> nstr:
        pass

    @abstractmethod
    def service_header(self, information: ServiceInformation) -> None:
        pass

    @abstractmethod
    def description_by_service_command(self, value: SC) -> ServiceDescription | None:
        pass

    @abstractmethod
    def error_detect_handler(
        exception: BaseException,
        host: nstr = None,
        argument_list: list[str] | None = None,
    ) -> None:
        pass

    @abstractmethod
    def service_was_not_started(
        self, service_information: ServiceInformation, error: str
    ) -> None:
        pass


class IServiceCollection(ABC):

    @abstractmethod
    def request(self, client) -> None:
        pass

    @abstractmethod
    def remove_service(self, value: ServiceDescriptionBase) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def get(
        self,
        key_description: ServiceDescription | None = None,
    ) -> dict[ServiceDescriptionBase, ServiceInformation] | ServiceInformation | None:
        pass

    @abstractmethod
    def update(
        self,
        value: ServiceDescriptionBase | list[ServiceDescriptionBase],
        add: bool = True,
        overwrite: bool = False,
    ) -> None:
        pass


class ServiceTool:

    @staticmethod
    def create_port(
        service_role_or_description: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> int:
        return (
            ServiceRoleTool.service_description(service_role_or_description).port
            or NetworkTool.next_free_port()
        )

    @staticmethod
    def create_host(
        service_role_or_description: SERVICE_ROLE | ServiceDescriptionBase,
    ) -> str:
        description: ServiceDescription = ServiceRoleTool.service_description(
            service_role_or_description
        )
        return (
            OSTool.host()
            if description.isolated or e(description.host)
            else description.host
        )

    @staticmethod
    def extract_parameter_list(value: ParameterList) -> ParameterList:
        return (
            ParameterList(value.get())
            if ServiceTool.has_subscribtion_result(value)
            else value
        )

    @staticmethod
    def create_event_listener_service_description(
        host: str, port: int
    ) -> ServiceDescription:
        return ServiceDescription(
            name=j(
                (EVENT_LISTENER_NAME_PREFIX, host, port),
                CONST.SPLITTER,
            ),
            description="Subscriber",
            host=host,
            port=port,
            auto_start=False,
            auto_restart=False,
        )

    @staticmethod
    def extract_service_information(
        value: dict | ServiceInformation,
    ) -> ServiceInformation:
        service_information: ServiceInformation = (
            DataTool.fill_data_from_source(ServiceInformation(), value)
            if isinstance(value, dict)
            else value
        )
        if ne(service_information.subscribtions):
            service_information.subscribtions = DataTool.fill_data_from_list_source(
                Subscribtion, service_information.subscribtions
            )
        return service_information

    @staticmethod
    def has_subscribtion_result(pl: ParameterList) -> bool:
        return (
            len(pl.values) == 2
            and DataTool.fill_data_from_source(
                SubscribtionResult(), pl.values[-1]
            ).checker
        )

    @staticmethod
    def is_service_as_listener(description: ServiceDescriptionBase) -> bool:
        return description.name.find(EVENT_LISTENER_NAME_PREFIX) == 0

    @staticmethod
    def is_service_as_support(description: ServiceDescriptionBase) -> bool:
        return description.name.find(SUPPORT_NAME_PREFIX) == 0


class ServiceRoleTool:

    @staticmethod
    def get(
        value: ServiceDescription | ServiceInformation | SERVICE_ROLE,
    ) -> SERVICE_ROLE | None:
        if isinstance(value, SERVICE_ROLE):
            return value
        for service_role in SERVICE_ROLE:
            if service_role.NAME.lower().replace("_", "") == value.name.lower():
                return service_role

    @staticmethod
    def service_description(
        value: Enum | str | ServiceDescriptionBase, get_source_description: bool = False
    ) -> ServiceDescriptionBase | None:
        def isolated_name(
            value: ServiceDescriptionBase | None,
        ) -> ServiceDescriptionBase | None:
            if n(value):
                return None
            value.name = (
                j((CONST.ISOLATED_ARG_NAME, value.name), CONST.SPLITTER)
                if value.isolated and value.name.find(CONST.ISOLATED_ARG_NAME) == -1
                else value.name
            )
            return value

        if isinstance(value, str):
            for service_role in SERVICE_ROLE:
                if ServiceRoleTool.service_description(service_role).name == value:
                    return isolated_name(service_role.value)
            return None
        if isinstance(value, ServiceDescriptionBase):
            return isolated_name(
                ServiceRoleTool.service_description(value.name)
                if get_source_description
                else value
            )
        return isolated_name(value.value)



import isgb

from enum import Enum
from sgb.collections import Host
from sgb.tools import StringTool, DataTool, first
from sgb.consts.community import COMMUNITY
from sgb.consts.server_role import SERVER_ROLE


class HOSTS(Enum):

    @property
    def NAME(self) -> str:
        return self.value.name

    @property
    def IP(self) -> str | None:
        return self.value.ip

    @property
    def ALIAS(self) -> str:
        return self.value.alias

    @staticmethod
    def get_by(host_name: str) -> Host | None:
        return first(
            DataTool.filter(
                lambda data: StringTool.contains(data.name, host_name, True),
                HOSTS,
            )
        )

    SPHERA_DC1 = Host("vld-np-dc1", "dc1")

    AD = SPHERA_DC1

    SPHERE_DC2 = Host("vld-np-dc2", "dc2")

    DEVELOPER = Host("vld-np-sgb-46")

    SKYPE = Host("vld-np-skype")

    SPHERE_PBX = Host("vld-np-pbx")

    HAPK_PBX = Host("krb-nv-pbx")

    PK_PBX = Host("vld-al-pbx")

    GAS_PBX = Host("trc-ml-pbx")

    AGMZ_PBX = Host("art-kr-pbx")

    MKUS_PBX = Host("us-mir-pbx")

    DAME_WARE = DEVELOPER

    EXECUTOR = DEVELOPER

    PASSWORD = DEVELOPER

    ROUTER = DEVELOPER

    SSH = DEVELOPER


class HOST_MAP:

    values: dict[HOSTS, tuple[SERVER_ROLE, COMMUNITY | tuple[COMMUNITY, ...]]] = {
        HOSTS.AGMZ_PBX: (SERVER_ROLE.TELEPHONY, (COMMUNITY.AGMZ, COMMUNITY.TD)),
        HOSTS.SPHERE_PBX: (SERVER_ROLE.TELEPHONY, COMMUNITY.SPHERE),
        HOSTS.HAPK_PBX: (SERVER_ROLE.TELEPHONY, COMMUNITY.HAPK),
        HOSTS.GAS_PBX: (SERVER_ROLE.TELEPHONY, COMMUNITY.GAS),
        HOSTS.PK_PBX: (SERVER_ROLE.TELEPHONY, COMMUNITY.PK),
        HOSTS.MKUS_PBX: (SERVER_ROLE.TELEPHONY, COMMUNITY.MKUS),
    }

    @staticmethod
    def by(server_role: SERVER_ROLE, community: COMMUNITY) -> Host | None:
        for host in HOST_MAP.values:
            value: tuple[SERVER_ROLE, COMMUNITY | tuple[COMMUNITY, ...]] = (
                HOST_MAP.values[host]
            )
            if (
                value[0] == server_role and community in value[1]
                if isinstance(value[1], tuple)
                else community == value[1]
            ):
                return host.value
        return None



from enum import Enum
from sgb.consts import CONST
from sgb.collections import (
    StorageVariableHolder,
    IntStorageVariableHolder,
    BoolStorageVariableHolder,
    TimeStorageVariableHolder,
    FloatStorageVariableHolder,
    IntListStorageVariableHolder,
    DateListStorageVariableHolder,
    StringListStorageVariableHolder,
    IntVariantableStorageVariableHolder,
)


class SETTINGS(Enum):

    USE_GOOGLE_KEEP = BoolStorageVariableHolder(
        "USE_GOOGLE_KEEP",
        False,
    )


    UNISENDER_API_KEY = StorageVariableHolder(
        "UNISENDER_API_KEY",
        "65ddrin3zh791hxarbkwe4fmah5p44hkg4cjwsuy",
    )

    UNISENDER_API_LIST_IDS = StorageVariableHolder(
        "UNISENDER_API_LIST_IDS",
        "761",
    )

    SGB_CLI_ADMINISTRATOR_LOGIN = StorageVariableHolder(
        "SGB_CLI_ADMINISTRATOR_LOGIN",
        "nak",
    )

    USER_RESPONSIBLE_FOR_PATIENT_MARK = StorageVariableHolder(
        "USER_RESPONSIBLE_FOR_PATIENT_MARK",
        "baa",
    )

    WIFI_VIP_PASSWORD = StorageVariableHolder("WIFI_VIP_PASSWORD", "ilovepacific")

    PLAIN_FORMAT_AS_DEFAULT_LOGIN_LIST = StringListStorageVariableHolder(
        "PLAIN_FORMAT_AS_DEFAULT_LOGIN_LIST", ("bar", "rob", "ptyu")
    )

    CHILLER_RECOGNIZE_LOG_LEVEL = IntStorageVariableHolder(
        "CHILLER_RECOGNIZE_LOG_LEVEL", 0
    )

    HEART_BEAT_IS_ON = BoolStorageVariableHolder("HEART_BEAT_IS_ON", True)

    CT_INDICATIONS_VALUE_TEMPERATURE_CORRECTION = FloatStorageVariableHolder(
        "CT_INDICATIONS_VALUE_TEMPERATURE_CORRECTION", 0.7
    )

    CT_INDICATIONS_VALUE_SAVE_PERIOD_IN_MINUTES = IntStorageVariableHolder(
        "CT_INDICATIONS_VALUE_SAVE_PERIOD_IN_MINUTES", 60
    )

    CHILLER_INDICATIONS_VALUE_SAVE_PERIOD_IN_MINUTES = IntStorageVariableHolder(
        "CHILLER_INDICATIONS_VALUE_SAVE_PERIOD_IN_MINUTES", 60
    )

    HOSPITAL_WORK_DAY_START_TIME = TimeStorageVariableHolder(
        "HOSPITAL_WORK_DAY_START_TIME", "8:00"
    )
    HOSPITAL_WORK_DAY_END_TIME = TimeStorageVariableHolder(
        "HOSPITAL_WORK_DAY_END_TIME", "20:00"
    )
    OFFICE_WORK_DAY_START_TIME = TimeStorageVariableHolder(
        "OFFICE_WORK_DAY_START_TIME", "8:30"
    )
    OFFICE_WORK_DAY_END_TIME = TimeStorageVariableHolder(
        "OFFICE_WORK_DAY_END_TIME", "17:00"
    )

    INDICATION_CT_NOTIFICATION_START_TIME = DateListStorageVariableHolder(
        "INDICATION_CT_NOTIFICATION_START_TIME", None
        #("8:00", "12:00", "15:00", "17:00")
    )

    USER_USE_CACHE = BoolStorageVariableHolder("USER_USE_CACHE", True)

    TIME_TRACKING_FOR_POLYCLINIC = StringListStorageVariableHolder(
        "TIME_TRACKING_FOR_POLYCLINIC",
        (
            "035",
            "045",
            "058",
            "101-0-",
            "124",
            "125",
            "131",
            "137",
            "139",
            #"177",
            "183",
            "190",
        ),
    )

    #
    RESOURCE_MANAGER_CHECK_SITE_CERTIFICATE_START_TIME = TimeStorageVariableHolder(
        "RESOURCE_MANAGER_CHECK_SITE_CERTIFICATE_START_TIME", "8:00"
    )

    #
    RESOURCE_MANAGER_CHECK_SITE_FREE_SPACE_PERIOD_IN_MINUTES = IntStorageVariableHolder(
        "RESOURCE_MANAGER_CHECK_SITE_FREE_SPACE_PERIOD_IN_MINUTES", 15
    )
    #
    PRINTER_REPORT_PERIOD_IN_MINUTES = IntStorageVariableHolder(
        "PRINTER_REPORT_PERIOD_IN_MINUTES", 5
    )

    PRINTER_MINIMAL_REPORT_VALUE = IntStorageVariableHolder(
        "PRINTER_MINIMAL_REPORT_VALUE", 10
    )

    OUTSORCE_PRINTER_REPORT_IS_ON = BoolStorageVariableHolder(
        "OUTSORCE_PRINTER_REPORT_IS_ON", True
    )

    OUTSORCE_PRINTER_REPORT_START_TIME = TimeStorageVariableHolder(
        "OUTSORCE_PRINTER_REPORT_TIME", "9:00"
    )
    #
  

    BONUS_PROGRAM_IS_ON = BoolStorageVariableHolder(
        "BONUS_PROGRAM_IS_ON",
        False,
    )

    BONUS_PROGRAM_DISCOUNT_PERCENT = IntStorageVariableHolder(
        "BONUS_PROGRAM_DISCOUNT_PERCENT",
        50,
    )

    BONUS_PROGRAM_CASHBACK_PERCENT = IntStorageVariableHolder(
        "BONUS_PROGRAM_CASHBACK_PERCENT",
        3,
    )

    BONUS_PROGRAM_BONUS_MINIMUM = IntStorageVariableHolder(
        "BONUS_PROGRAM_BONUS_MINIMUM",
        10,
    )

    BONUS_PROGRAM_NOTIFICATION_COUNT_MAX = IntStorageVariableHolder(
        "BONUS_PROGRAM_NOTIFICATION_COUNT_MAX",
        3,
    )

    WHATSAPP_SENDING_MESSAGES_VIA_WAPPI_IS_ON = BoolStorageVariableHolder(
        "WHATSAPP_SENDING_MESSAGES_VIA_WAPPI_IS_ON", True
    )

    WHATSAPP_BUFFERED_MESSAGE_MIN_DELAY_IN_MILLISECONDS = IntStorageVariableHolder(
        "WHATSAPP_BUFFERED_MESSAGE_MIN_DELAY_IN_MILLISECONDS", 6000
    )

    WHATSAPP_BUFFERED_MESSAGE_MAX_DELAY_IN_MILLISECONDS = IntStorageVariableHolder(
        "WHATSAPP_BUFFERED_MESSAGE_MAX_DELAY_IN_MILLISECONDS", 12000
    )

    WHATSAPP_MESSAGE_SENDER_USER_LOGIN = StorageVariableHolder(
        "WHATSAPP_MESSAGE_SENDER_USER_LOGIN", "Administrator"
    )

    WORKSTATION_SHUTDOWN_TIME = TimeStorageVariableHolder(
        "WORKSTATION_SHUTDOWN_TIME", "21:00"
    )

    WORKSTATION_REBOOT_TIME = TimeStorageVariableHolder(
        "WORKSTATION_REBOOT_TIME", "21:00"
    )

    EMAIL_VALIDATION_IS_ON = BoolStorageVariableHolder("EMAIL_VALIDATION_IS_ON", True)
    EMAIL_VALIDATION_TEST = BoolStorageVariableHolder("EMAIL_VALIDATION_TEST", False)

    CHILLER_ALERT_TEMPERATURE = FloatStorageVariableHolder(
        "CHILLER_ALERT_TEMPERATURE", 17.0
    )

    CHILLER_COUNT_DEFAULT = FloatStorageVariableHolder("CHILLER_COUNT_DEFAULT", 3)

    CHILLER_MAX_TEMPERATURE = IntStorageVariableHolder("CHILLER_MAX_TEMPERATURE", 16)

    CHILLER_MIN_TEMPERATURE = IntStorageVariableHolder("CHILLER_MIN_TEMPERATURE", 10)

    CHILLER_ACTION_MAX_TEMPERATURE = FloatStorageVariableHolder(
        "CHILLER_ACTION_MAX_TEMPERATURE", 15.1
    )

    CHILLER_ACTION_MIN_TEMPERATURE = IntStorageVariableHolder(
        "CHILLER_ACTION_MIN_TEMPERATURE", 11
    )

    CT_ROOM_MAX_TEMPERATURE = IntStorageVariableHolder("CT_ROOM_MAX_TEMPERATURE", 24)

    CT_ROOM_MIN_TEMPERATURE = IntStorageVariableHolder("CT_ROOM_MIN_TEMPERATURE", 20)

    CT_ROOM_MAX_HUMIDITY = IntStorageVariableHolder("CT_ROOM_MAX_HUMIDITY", 60)

    CT_ROOM_MIN_HUMIDITY = IntStorageVariableHolder("CT_ROOM_MIN_HUMIDITY", 30)

    CHECK_ALL_RECIPIENT_USER_LOGIN = StorageVariableHolder(
        "CHECK_ALL_RECIPIENT_USER_LOGIN", "nak"
    )

    MAIL_SITE_FREE_MEMORY_PERIOD = IntStorageVariableHolder(
        "MAIL_SITE_FREE_MEMORY_PERIOD", 180
    )

    SITE_FREE_MEMORY_PERIOD = IntStorageVariableHolder("SITE_FREE_MEMORY_PERIOD", 180)



import isgb

# from sgb.consts.hosts import HOSTS
from sgb.collections import nstr, nint, nbool, strtuple, Host

from dataclasses import dataclass, field
from typing import Any
from enum import Enum


@dataclass
class ServiceDescriptionBase:
    name: nstr = None
    host: nstr = None
    port: nint = None
    service_path: nstr = None
    isolated: bool = False
    host_changeable: bool = True
    visible_for_admin: bool = True
    auto_start: bool = True
    auto_restart: bool = True
    run_from_system_account: bool = False
    python_executable_path: nstr = None
    version: nstr = None
    sgb_version: nstr = None
    parameters: Any = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, another) -> bool:
        return (
            False
            if another is None
            else (
                self.name == another.name
                if isinstance(another, ServiceDescriptionBase)
                else self.name == another
            )
        )


@dataclass
class SubscribtionDescription:
    service_name: nstr = None
    service_command: Enum | str | None = None
    type: nint = None
    name: nstr = None


@dataclass
class Subscribtion(SubscribtionDescription):
    available: bool = False
    enabled: bool = False


@dataclass
class SubscribtionInformation(SubscribtionDescription):
    pass


@dataclass
class ServiceInformation(ServiceDescriptionBase):
    subscribtions: list[Subscribtion] = field(default_factory=list)
    pid: int = -1
    standalone: nbool = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, another: Any):
        return (
            False
            if another is None
            else (
                self.name == another.name
                if isinstance(another, ServiceDescriptionBase)
                else self.name == another
            )
        )


@dataclass
class ServiceDescription(ServiceDescriptionBase):
    description: nstr = None
    login: nstr = None
    password: nstr = None
    commands: tuple[Enum | str, ...] | None = None
    use_standalone: bool = False
    standalone_name: nstr = None
    support_servers: strtuple | None = None
    packages: strtuple | None = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, another):
        return (
            False
            if another is None
            else (
                self.name == another.name
                if isinstance(another, ServiceDescriptionBase)
                else self.name == another
            )
        )


@dataclass
class SubscriberInformation:
    type: nint = None
    name: nstr = None
    available: bool = True
    enabled: bool = True
    service_information: ServiceDescriptionBase | None = None


@dataclass
class SubscribtionResult:
    result: Any = None
    type: int = 0
    checker: bool = False


class ServiceRoleItem(Enum):

    @property
    def value(self) -> ServiceDescription:
        return super().value

    @property
    def host(self) -> str | None:
        return self.value.host

    @property
    def standalone_name(self) -> str | None:
        return self.value.standalone_name

    @host.setter
    def host(self, value: str | Host) -> None:
        if isinstance(value, str):
            self.value.host = value
        elif isinstance(value, Host):
            self.host = value.value

    @property
    def NAME(self) -> str | None:
        return self.value.name



import isgb
from sgb.consts.community import COMMUNITY
"""
trc-ml-pbx	192.168.202.8	1XXX	ООО «Грин Агро-Сахалин»	11xx	-
trc-ml-pbx	192.168.202.8	1XXX	ООО ТД «Дальневосточные Продукты»	13xx	-
vld-np-pbx	192.168.2.8	2XXX	Грин Агро Нижнепортовая\ООО «ХАПК «Грин Агро»	21xx	-
vld-np-pbx	192.168.2.8	2XXX	ООО «Сфера Менеджмент»	22xx	-
vld-np-pbx	192.168.2.8	2XXX	ООО «Грин Агро-Кировский»	29xx	-
vld-np-pbx	192.168.2.8	2XXX	ООО ТД «Дальневосточные Продукты»	28xx	-
art-kr-pbx	192.168.1.9	3XXX	АО ГМЗ «Артёмовский»	31xx	-
art-kr-pbx	192.168.1.9	3XXX	ООО ТД «Дальневосточные Продукты»	33xx	-
krb-nv-pbx	192.168.102.9	4XXX	ООО «ХАПК «Грин Агро»	41xx	В перспективе
us-mir-pbx	192.168.205.11	5XXX	ОАО «Молочный комбинат «Южно-Сахалинский»	52xx	54хх - телефоны операторов (заявки)
-	-	6XXX	-	61xx	Свободно
vld-al-pbx	192.168.0.5	7XXX	ООО «Приморский кондитер» Руднева	71xx-78xx	-
"""

TELEPHONE_POOL: dict[COMMUNITY, tuple[dict[COMMUNITY, tuple[int, int]], ...]] = {
    COMMUNITY.SPHERE: {
        COMMUNITY.SPHERE: (2200, 2299),
        COMMUNITY.TD: (2800, 2899),
        COMMUNITY.GAS: (2100, 2199),
        COMMUNITY.HAPK: (2100, 2199),
        COMMUNITY.GAK: (2900, 2999),
    },
    COMMUNITY.GAS: {COMMUNITY.GAS: (1100, 1199), COMMUNITY.TD: (1300, 1399)},
    COMMUNITY.PK: {COMMUNITY.PK: (7000, 7899)},
    COMMUNITY.AGMZ: {COMMUNITY.AGMZ: (3100, 3199), COMMUNITY.TD: (3300, 3399)},
    COMMUNITY.HAPK: {COMMUNITY.HAPK: (4100, 4199)},
    COMMUNITY.MKUS: {COMMUNITY.MKUS: (5000, 5999)},
}



import isgb

from sgb.collections import nstr, nbool
from sgb.tools import j_s, escs, n

from dataclasses import dataclass

from typing import Any

import grpc


RPCError = grpc.RpcError


class NotImplemented(BaseException):
    pass


class OperationExit(BaseException):
    pass


class OperationCanceled(BaseException):
    pass


@dataclass
class Redirection(BaseException):
    arg: Any | None = None


class ZeroReached(BaseException):
    pass


class BarcodeNotFound(BaseException):

    def get_details(self) -> str:
        return "Штрих-код не распознан, попробуйте еще раз"


class NotFound(BaseException):

    def get_details(self) -> str:
        return self.args[0]

    def get_value(self) -> str:
        return self.args[1]


class IncorrectInputFile(BaseException):
    pass


class NotAccesable(BaseException):
    pass


@dataclass
class Error(BaseException):
    details: nstr = None
    code: tuple | None = None

class ServiceIsNotStartedError(Error):
    pass

class ERROR:

    class USER:

        @staticmethod
        def create_not_found_error(title: str, active: nbool, value: str) -> NotFound:
            start: nstr = None
            if n(active):
                start = "Пользователь"
            elif active:
                start = "Активный пользователь"
            else:
                start = "Неактивный пользователь"
            return NotFound(j_s((start, "с", title, escs(value), "не найден!")), value)

    class WORKSTATION:

        @staticmethod
        def create_not_found_error(title: str, value: str) -> NotFound:
            return NotFound(
                j_s(("Компьютер", "с", title, escs(value), "не найден!")), value
            )
